/*
 * Renderer 7. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.models_L.Disk;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This example demonstrates "cropping" combined with "scaling"
   (and no letter boxing).
<p>
   This program always uses a viewport that fills the framebuffer.
   The aspect ratio of the viewport is always the same as the
   framebuffer's. In order not to distort the image from the view-
   rectangle, this program must match the view-rectangle's aspect
   ratio to the viewport (framebuffer) aspect ratio. But if that
   aspect ratio is not 1, then the view-rectangle cannot enclose
   all of the disk from the view-plane. In that case, some of the
   disk must be cropped off by the view-rectangle.
<p>
   Notice that this program never uses letter boxing. Since the
   viewport always fills up the framebuffer, there is never any
   empty framebuffer space to letterbox.
<p>
   If the program window has an aspect ration that is greater than
   1, then this program will scale the view-rectangle's x-axis to
   stretch horizontally all the way across that viewport (framebuffer).
   This program will then need to crop the view-rectangle along the
   y-axis.
<p>
   If the program window has an aspect ration that is less than
   1, then this program will scale the view-rectangle's y-axis to
   stretch vertically all the way across that viewport (framebuffer).
   This program will then need to crop the view-rectangle along the
   x-axis.
<p>
   This program has nine modes for how it anchors the circle
   in the camera's view-rectangle.
<p>
   Mode 1 anchors the circle in the upper left-hand corner
   of the view-rectangle.
<p>
   Mode 2 anchors the circle at the center of the top edge
   of the view-rectangle.
<p>
   Mode 3 anchors the circle in the upper right-hand corner
   of the view-rectangle.
<p>
   Mode 4 anchors the circle at the center of the right edge
   of the view-rectangle.
<p>
   Mode 5 anchors the circle in the lower right-hand corner
   of the view-rectangle.
<p>
   Mode 6 anchors the circle at the center of the bottom edge
   of the framebuffer.
<p>
   Mode 7 anchors the circle in the lower left-hand corner
   of the view-rectangle.
<p>
   Mode 8 anchors the circle at the center of the left edge
   of the view-rectangle.
<p>
   Mode 9 anchors the circle at the center of the view-rectangle.
<p>
  As the user changes the shape (the aspect ratio) and size of
  the program window, the code in the componentResized() event
  handler needs to change the size and location of the
  framebuffer's viewport (within the framebuffer) and also the
  size and location of the camera's view volume (within view
  space).
<p>
  We create this effect in the setupViewing() method (called
  by the componentResized() event handler) by using the
  FrameBuffer's setViewport() method and the Camera's
  projOrtho() method.
*/
@SuppressWarnings("serial")
public class Circle_v3_Crop_Scale extends Circle_v0_Abstract
{
   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public Circle_v3_Crop_Scale()
   {
      super("Renderer 7 - Crop & Scale");
      print_help_message();
   }


   /**
      Get in one place the code to set up
      the view volume and the viewport.
   */
   @Override protected void setupViewing()
   {
      // Get the size of the FrameBuffer.
      final FrameBuffer fb = fbp.getFrameBuffer();
      final int wFB = fb.width;
      final int hFB = fb.height;

      // Make the viewport all of the FrameBuffer.
      final int wVP = wFB;
      final int hVP = hFB;
      fb.setViewport(0, 0, wVP, hVP);

      // Compute the viewport, framebuffer, view-rectangle aspect ratio.
      final double r = (double)wVP / (double)hVP;

      if (mode == 1)
      {
         // 1. upper left-hand corner
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             -1,                     // left
                             (r >= 1) ? +1 : 2*r-1,  // right
                             (r >= 1) ? 1-2/r : -1,  // bottom
                             +1));                   // top
      }
      else if (mode == 2)
      {
         // 2. center of the top edge
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1 : -r,     // left
                             (r >= 1) ? +1 : +r,     // right
                             (r >= 1) ? 1-2/r : -1,  // bottom
                             +1));                   // top
      }
      else if (mode == 3)
      {
         // 3. upper right-hand corner
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1 : 1-2*r,  // left
                             +1,                     // right
                             (r >= 1) ? 1-2/r : -1,  // bottom
                             +1));                   // top
      }
      else if (mode == 4)
      {
         // 4. center of the right edge
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1 : 1-2*r,  // left
                             +1,                     // right
                             (r >= 1) ? -1/r : -1,   // bottom
                             (r >= 1) ? +1/r : +1)); // top
      }
      else if (mode == 5)
      {
         // 5. lower right-hand corner
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1 : 1-2*r,   // left
                             +1,                      // right
                             -1,                      // bottom
                             (r >= 1) ? 2/r-1 : +1)); // top
      }
      else if (mode == 6)
      {
         // 6. center of the bottom edge
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1 : -r,      // left
                             (r >= 1) ? +1 : +r,      // right
                             -1,                      // bottom
                             (r >= 1) ? 2/r-1 : +1)); // top
      }
      else if (mode == 7)
      {
         // 7. lower left-hand corner
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             -1,                      // left
                             (r >= 1) ? +1 : 2*r-1,   // right
                             -1,                      // bottom
                             (r >= 1) ? 2/r-1 : +1)); // top
      }
      else if (mode == 8)
      {
         // 8. centered of the left edge
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             -1,                     // left
                             (r >= 1) ? +1 : 2*r-1,  // right
                             (r >= 1) ? -1/r : -1,   // bottom
                             (r >= 1) ? +1/r : +1)); // top
      }
      else if (mode == 9)
      {
         // 9. center of the framebuffer
         scene = scene.changeCamera(
                     Camera.projOrtho(
                             (r >= 1) ? -1   : -r,   // left
                             (r >= 1) ? +1   : +r,   // right
                             (r >= 1) ? -1/r : -1,   // bottom
                             (r >= 1) ? +1/r : +1)); // top
      }

      super.setupViewing();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Circle_v3_Crop_Scale()
      );
   }
}
