/*
 * Renderer 7. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our sixth rendering pipeline. It adds a view
   volume normalization stage, {@link View2Camera}, that converts
   a {@link Camera}'s configurable view volume into the normalized
   view volume used by the {@link Clip} pipeline stage. Coordinates
   relative to the {@link Camera}'s arbitrary view volume are called
   "view coordinates" and coordinates relative to the normalized view
   volume are called "camera coordinates", so the new pipeline stage
   converts vertex coordinates from view coordinates to camera
   coordinates. There are six pipeline stages.
*/
public final class Pipeline
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n== Begin Rendering of Scene: " + scene.name + " ==");

      logMessage("-- Current Camera:\n" + scene.camera);

      // For every Position in the Scene, render the Position's Model.
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            logMessage("==== Render position: " + position.name + " ====");

            logMessage("---- Translation vector = " + position.getTranslation());

            if ( position.getModel().visible )
            {
               logMessage("====== Render model: "
                                  + position.getModel().name + " ======");

               CheckModel.check(position.getModel());

               // Mostly for compatibility with renderers 1 through 3.
               if (  position.getModel().colorList.isEmpty()
                 && !position.getModel().vertexList.isEmpty())
               {
                  for (int i = 0; i < position.getModel().vertexList.size(); ++i)
                  {
                     position.getModel().addColor( DEFAULT_COLOR );
                  }
                  System.err.println("***WARNING: Added default color to model: "
                                    + position.getModel().name + ".");
               }

               logVertexList("0. Model       ", position.getModel());

               // 1. Apply the Position's model-to-view coordinate transformation.
               final Model model1 = Model2View.model2view(position);

               logVertexList("1. View        ", model1);

               // 2. Apply the Camera's normalizing view-to-camera coordinate transformation.
               final Model model2 = View2Camera.view2camera(model1,
                                                            scene.camera);

               logVertexList("2. Camera      ", model2);
               logColorList("2. Camera      ", model2);
               logPrimitiveList("2. Camera      ", model2);

               // 3. Clip primitives to the camera's near plane.
               final Model model3 = NearClip.clip(model2,
                                                  scene.camera);

               logVertexList("3. Near_Clipped", model3);
               logColorList("3. Near_Clipped", model3);
               logPrimitiveList("3. Near_Clipped", model3);

               // 4. Apply the Camera's projection transformation.
               final Model model4 = Projection.project(model3,
                                                       scene.camera);

               logVertexList("4. Projected   ", model4);

               // 5. Clip primitives to the camera's view rectangle.
               final Model model5 = Clip.clip(model4);

               logVertexList("5. Clipped     ", model5);
               logColorList("5. Clipped     ", model5);
               logPrimitiveList("5. Clipped     ", model5);

               // 6. Rasterize every visible primitive into pixels.
               Rasterize.rasterize(model5, vp);

               logMessage("====== End model: "
                                  + position.getModel().name + " ======");
            }
            else
            {
               logMessage("====== Hidden model: "
                                  + position.getModel().name + " ======");
            }

            logMessage("==== End position: " + position.name + " ====");
         }
         else
         {
            logMessage("==== Hidden position: " + position.name + " ====");
         }
      }
      logMessage("== End Rendering of Scene ==");
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline() {
      throw new AssertionError();
   }
}
