/*
 * Renderer 7. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.primitives.LineSegment;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   Rasterize a clipped {@link LineSegment} into shaded pixels
   in a {@link FrameBuffer}'s viewport and (optionally)
   anti-alias and gamma-encode the line at the same time.
<p>
   This pipeline stage takes a clipped {@link LineSegment}
   with vertices in the {@link Camera}'s view rectangle and
   rasterizezs the line segment into shaded, anti-aliased
   pixels in a {@link FrameBuffer}'s viewport. This rasterizer
   will linearly interpolate color from the line segment's two
   endpoints to each rasterized (and anti-aliased) pixel in
   the line segment.
<p>
   This rasterization algorithm is based on
<pre>
     "Fundamentals of Computer Graphics", 3rd Edition,
      by Peter Shirley, pages 163-165.
</pre>
<p>
   This rasterizer implements a simple version of Xiaolin_Wu's
   anti-aliasing algorithm. See
     <a href="https://en.wikipedia.org/wiki/Xiaolin_Wu's_line_algorithm" target="_top">
              https://en.wikipedia.org/wiki/Xiaolin_Wu's_line_algorithm</a>
*/
public class Rasterize_AntiAlias_Line
{
   /**
      Rasterize a clipped {@link LineSegment} into anti-aliased, shaded pixels
      in the {@link FrameBuffer.Viewport}.

      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} to rasterize into the {@link FrameBuffer.Viewport}
      @param vp     {@link FrameBuffer.Viewport} to hold rasterized pixels
   */
   public static void rasterize(final Model model,
                                final LineSegment ls,
                                final FrameBuffer.Viewport vp)
   {
      // Get the viewport's background color.
      final Color bg = vp.bgColorVP;

      // Make local copies of several values.
      final int w = vp.getWidthVP();
      final int h = vp.getHeightVP();

      final int vIndex0 = ls.vIndexList.get(0);
      final int vIndex1 = ls.vIndexList.get(1);
      final Vertex v0 = model.vertexList.get(vIndex0);
      final Vertex v1 = model.vertexList.get(vIndex1);

      final int cIndex0 = ls.cIndexList.get(0);
      final int cIndex1 = ls.cIndexList.get(1);
      final float[] c0 = model.colorList.get(cIndex0).getRGBComponents(null);
      final float[] c1 = model.colorList.get(cIndex1).getRGBComponents(null);
      float r0 = c0[0], g0 = c0[1], b0 = c0[2];
      float r1 = c1[0], g1 = c1[1], b1 = c1[2];

      // Transform each vertex to the "pixel plane" coordinate system.
      double x0 = 0.5 + w/2.001 * (v0.x + 1); // x_pp = 0.5 + w/2 * (x_p+1)
      double y0 = 0.5 + h/2.001 * (v0.y + 1); // y_pp = 0.5 + h/2 * (y_p+1)
      double x1 = 0.5 + w/2.001 * (v1.x + 1);
      double y1 = 0.5 + h/2.001 * (v1.y + 1);
      // NOTE: Notice the 2.001 fudge factor in the last two equations.
      // This is explained on page 142 of
      //    "Jim Blinn's Corner: A Trip Down The Graphics Pipeline"
      //     by Jim Blinn, 1996, Morgan Kaufmann Publishers.

      if (Rasterize.debug)
      {
         logMessage(String.format("(x0_pp, y0_pp) = (%9.4f, %9.4f)", x0,y0));
         logMessage(String.format("(x1_pp, y1_pp) = (%9.4f, %9.4f)", x1,y1));
      }

      // Round each vertex to the nearest logical pixel.
      // This makes the algorithm a lot simpler, but it can
      // cause a slight, but noticeable, shift of the line segment.
      x0 = Math.round(x0);
      y0 = Math.round(y0);
      x1 = Math.round(x1);
      y1 = Math.round(y1);

      // Rasterize a degenerate line segment (a line segment
      // that projected onto a point) as a single pixel.
      if ( (x0 == x1) && (y0 == y1) )
      {
         // We don't know which endpoint of the line segment
         // is in front, so just pick v0.
         final int x0_vp = (int)x0 - 1;  // viewport coordinate
         final int y0_vp = h - (int)y0;  // viewport coordinate

         if (Rasterize.debug) logPixel(x0, y0,
                                       x0_vp, y0_vp,
                                       r0, g0, b0,
                                       vp);
         // Log the pixel before setting it so that an array out-
         // of-bounds error will be right after the pixel's address.

         vp.setPixelVP(x0_vp, y0_vp, new Color(r0, g0, b0));

         return;
      }

      // If abs(slope) > 1, then transpose this line so that
      // the transposed line has slope < 1. Remember that the
      // line has been transposed.
      boolean transposedLine = false;
      if (Math.abs(y1 - y0) > Math.abs(x1 - x0)) // if abs(slope) > 1
      {
         // Swap x0 with y0.
         final double temp0 = x0;
         x0 = y0;
         y0 = temp0;
         // Swap x1 with y1.
         final double temp1 = x1;
         x1 = y1;
         y1 = temp1;
         transposedLine = true; // Remember that this line is transposed.
      }

      if (x1 < x0) // We want to rasterize in the direction of increasing x,
      {            // so, if necessary, swap (x0, y0) with (x1, y1).
         // Swap x0 with x1.
         final double tempX = x0;
         x0 = x1;
         x1 = tempX;
         // Swap y0 with y1.
         final double tempY = y0;
         y0 = y1;
         y1 = tempY;
         // Swap the colors too.
         final float tempR = r0;
         final float tempG = g0;
         final float tempB = b0;
         r0 = r1;
         g0 = g1;
         b0 = b1;
         r1 = tempR;
         g1 = tempG;
         b1 = tempB;
      }

      // Compute this line segment's slopes.
      final double      m = (y1 - y0) / (x1 - x0);
      final double slopeR = (r1 - r0) / (x1 - x0);
      final double slopeG = (g1 - g0) / (x1 - x0);
      final double slopeB = (b1 - b0) / (x1 - x0);

      if (Rasterize.debug)
      {
         final String inverseSlope = (transposedLine)
                                        ? " (transposed, so 1/m = " + 1/m + ")"
                                        : "";
         logMessage("Slope m    = " + m + inverseSlope);
         logMessage("Slope mRed = " + slopeR);
         logMessage("Slope mGrn = " + slopeG);
         logMessage("Slope mBlu = " + slopeB);
         logMessage(String.format("(x0_vp, y0_vp) = (%9.4f, %9.4f)", x0-1,h-y0));
         logMessage(String.format("(x1_vp, y1_vp) = (%9.4f, %9.4f)", x1-1,h-y1));
      }

      // Rasterize this line segment in the direction of increasing x.
      // In the following loop, as x moves across the logical horizontal
      // (or vertical) pixels, we will compute a y value for each x.
      double y = y0;
      for (int x = (int)x0; x < (int)x1; x += 1, y += m)
      {
         // Interpolate this pixel's color between the two endpoint's colors.
         float r = (float)Math.abs(r0 + slopeR*(x - x0));
         float g = (float)Math.abs(g0 + slopeG*(x - x0));
         float b = (float)Math.abs(b0 + slopeB*(x - x0));
         // We need the Math.abs() because otherwise, we sometimes get -0.0.

         if (Rasterize.doAntiAliasing)
         {
            // y must be between two vertical (or horizontal) logical pixel
            //  coordinates. Let y_low and y_hi be the logical pixel coordinates
            // that bracket around y.
            int y_low = (int)y;                      // the integer part of y
            int y_hi  = y_low + 1;
            if (!transposedLine && y == h) y_hi = h; // test for the top edge
            if ( transposedLine && y == w) y_hi = w; // test for the right edge

            // Let weight be the fractional part of y. We will use
            // weight to determine how much emphasis to place on
            // each of the two pixels that bracket y.
            final float weight = (float)(y - y_low);

            // Interpolate colors for the low and high pixels.
            // The smaller weight is, the closer y is to the lower
            // pixel, so we give the lower pixel more emphasis when
            // weight is small.
            float r_low = (1 - weight) * r + weight * (bg.getRed()  /255.0f);
            float g_low = (1 - weight) * g + weight * (bg.getGreen()/255.0f);
            float b_low = (1 - weight) * b + weight * (bg.getBlue() /255.0f);
            float r_hi  = weight * r + (1 - weight) * (bg.getRed()  /255.0f);
            float g_hi  = weight * g + (1 - weight) * (bg.getGreen()/255.0f);
            float b_hi  = weight * b + (1 - weight) * (bg.getBlue() /255.0f);
/*
            // You can try replacing the above anti-aliasing code with this
            // code to see that this simple idea doesn't work here (as it
            // did in the previous renderer). This code just distributes the
            // line's color between two adjacent pixels (instead of blending
            // each pixel's color with the back ground color). This code ends
            // up having pixels fade to black, instead of fading to the back
            // ground color.
            float r_low = (1 - weight) * r;
            float g_low = (1 - weight) * g;
            float b_low = (1 - weight) * b;
            float r_hi  = weight * r;
            float g_hi  = weight * g;
            float b_hi  = weight * b;
*/
            if (Rasterize.doGamma)
            {
               // Apply gamma-encoding (gamma-compression) to the colors.
               // https://www.scratchapixel.com/lessons/digital-imaging/digital-images
               // http://blog.johnnovak.net/2016/09/21/what-every-coder-should-know-about-gamma/
               final double gammaInv = 1.0 / Rasterize.GAMMA;
               r_low = (float)Math.pow(r_low, gammaInv);
               r_hi  = (float)Math.pow(r_hi,  gammaInv);
               g_low = (float)Math.pow(g_low, gammaInv);
               g_hi  = (float)Math.pow(g_hi,  gammaInv);
               b_low = (float)Math.pow(b_low, gammaInv);
               b_hi  = (float)Math.pow(b_hi,  gammaInv);
            }

            // Set this (anti-aliased) pixel in the framebuffer.
            if ( ! transposedLine )
            {
               final int x_vp     = x - 1;      // viewport coordinate
               final int y_vp_low = h - y_low;  // viewport coordinate
               final int y_vp_hi  = h - y_hi;   // viewport coordinate

               if (Rasterize.debug) logPixelsAA(x, y,
                                                x_vp, y_vp_low, y_vp_hi,
                                                r_low, g_low, b_low,
                                                r_hi,  g_hi,  b_hi,
                                                vp);

               vp.setPixelVP(x_vp, y_vp_low, new Color(r_low, g_low, b_low));
               vp.setPixelVP(x_vp, y_vp_hi,  new Color(r_hi,  g_hi,  b_hi));
            }
            else // a transposed line
            {
               final int x_vp_low = y_low - 1;  // viewport coordinate
               final int x_vp_hi  = y_hi  - 1;  // viewport coordinate
               final int y_vp     = h - x;      // viewport coordinate

               if (Rasterize.debug) logPixelsAA(y, x,
                                                x_vp_low, x_vp_hi, y_vp,
                                                r_low, g_low, b_low,
                                                r_hi,  g_hi,  b_hi,
                                                vp);

               vp.setPixelVP(x_vp_low, y_vp, new Color(r_low, g_low, b_low));
               vp.setPixelVP(x_vp_hi,  y_vp, new Color(r_hi,  g_hi,  b_hi));
            }
         }
         else // No anti-aliasing.
         {
            if (Rasterize.doGamma)
            {
               // Apply gamma-encoding (gamma-compression) to the colors.
               // https://www.scratchapixel.com/lessons/digital-imaging/digital-images
               // http://blog.johnnovak.net/2016/09/21/what-every-coder-should-know-about-gamma/
               final double gammaInv = 1.0 / Rasterize.GAMMA;
               r = (float)Math.pow(r, gammaInv);
               g = (float)Math.pow(g, gammaInv);
               b = (float)Math.pow(b, gammaInv);
            }

            // The value of y will almost always be between
            // two vertical (or horizontal) pixel coordinates.
            // By rounding off the value of y, we are choosing the
            // nearest logical vertical (or horizontal) pixel coordinate.
            if ( ! transposedLine )
            {
               final int x_vp = x - 1;                  // viewport coordinate
               final int y_vp = h - (int)Math.round(y); // viewport coordinate

               if (Rasterize.debug) logPixel(x, y,
                                             x_vp, y_vp,
                                             r, g, b,
                                             vp);
               // Log the pixel before setting it so that an array out-
               // of-bounds error will be right after the pixel's address.

               vp.setPixelVP(x_vp, y_vp, new Color(r, g, b));
            }
            else // a transposed line
            {
               final int x_vp = (int)Math.round(y) - 1; // viewport coordinate
               final int y_vp = h - x;                  // viewport coordinate

               if (Rasterize.debug) logPixel(y, x,
                                             x_vp, y_vp,
                                             r, g, b,
                                             vp);

               vp.setPixelVP(x_vp, y_vp, new Color(r, g, b));
            }
         }
         // Advance (x,y) to the next pixel (delta_x is 1, so delta_y is m).
      }
      // Set the pixel for the (x1,y1) endpoint.
      // We do this separately to avoid rounding errors.
      if ( ! transposedLine )
      {
         final int x_vp = (int)x1 - 1;  // viewport coordinate
         final int y_vp = h - (int)y1;  // viewport coordinate

         if (Rasterize.debug) logPixel(x1, y1,
                                       x_vp, y_vp,
                                       r1, g1, b1,
                                       vp);

         vp.setPixelVP(x_vp, y_vp, new Color(r1, g1, b1));
      }
      else // a transposed line
      {
         final int x_vp = (int)y1 - 1;  // viewport coordinate
         final int y_vp = h - (int)x1;  // viewport coordinate

         if (Rasterize.debug) logPixel(y1, x1,
                                       x_vp, y_vp,
                                       r1, g1, b1,
                                       vp);

         vp.setPixelVP(x_vp, y_vp, new Color(r1, g1, b1));
      }
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Rasterize_AntiAlias_Line() {
      throw new AssertionError();
   }
}
